import os
import random
import sys
from typing import List, Tuple

import wasmtime as wasm

from promptl_ai.rpc.errors import RPCError
from promptl_ai.rpc.payloads import Parameters
from promptl_ai.rpc.types import Call, Error, ErrorCode, Procedure, Result, _Calls, _Results
from promptl_ai.util import Model


class ClientOptions(Model):
    module_path: str
    working_dir: str


class Client:
    options: ClientOptions

    engine: wasm.Engine
    linker: wasm.Linker
    store: wasm.Store
    module: wasm.Module

    def __init__(self, options: ClientOptions):
        self.options = options

        config = wasm.Config()
        self.engine = wasm.Engine(config)
        self.linker = wasm.Linker(self.engine)
        self.linker.define_wasi()
        self.store = wasm.Store(self.engine)

        with open(self.options.module_path, "rb") as file:
            raw_module = file.read()
        wasm.Module.validate(self.engine, raw_module)
        self.module = wasm.Module(self.engine, raw_module)

    # Note: this is needed to avoid race condition on the file system
    # because we cannot reuse the same file io for multiple calls yet
    def _prepare(self) -> Tuple[str, str, str]:
        execution_id = str(random.randint(0, sys.maxsize))
        execution_dir = os.path.join(self.options.working_dir, execution_id)
        os.makedirs(execution_dir, exist_ok=True)

        stdin = os.path.join(execution_dir, "stdin")
        stdout = os.path.join(execution_dir, "stdout")
        stderr = os.path.join(execution_dir, "stderr")

        return stdin, stdout, stderr

    def _instantiate(self, stdin: str, stdout: str, stderr: str) -> wasm.Instance:
        config = wasm.WasiConfig()
        config.argv = []
        config.env = {}.items()
        config.stdin_file = stdin
        config.stdout_file = stdout
        config.stderr_file = stderr
        self.store.set_wasi(config)

        return self.linker.instantiate(self.store, self.module)

    def _send(self, stdin: str, data: bytes):
        with open(stdin, "wb") as file:
            file.write(data + b"\n")

    def _receive(self, stdout: str, stderr: str) -> Tuple[bytes, bytes]:
        with open(stdout, "rb") as file:
            out = file.read().strip()

        with open(stderr, "rb") as file:
            err = file.read().strip()

        return out, err

    def _execute(self, calls: List[Call]) -> List[Result]:
        stdin, stdout, stderr = self._prepare()

        try:
            self._send(stdin, _Calls.dump_json(calls))
        except Exception as exception:
            raise RPCError(
                Error(
                    code=ErrorCode.ReceiveError,
                    message=str(exception),
                )
            ) from exception

        try:
            instance = self._instantiate(stdin, stdout, stderr)
            instance.exports(self.store)["_start"](self.store)  # pyright: ignore [reportCallIssue]
        except Exception as exception:
            raise RPCError(
                Error(
                    code=ErrorCode.ExecuteError,
                    message=str(exception),
                )
            ) from exception

        out, err = self._receive(stdout, stderr)
        if err:
            raise RPCError(
                Error(
                    code=ErrorCode.UnknownError,
                    message=err.decode(),
                )
            )

        if not out:
            raise RPCError(
                Error(
                    code=ErrorCode.SendError,
                    message="No output",
                )
            )

        try:
            results = _Results.validate_json(out)
        except Exception as exception:
            raise RPCError(
                Error(
                    code=ErrorCode.SendError,
                    message=str(exception),
                )
            ) from exception

        return results

    def execute(self, procedure: Procedure, parameters: Parameters) -> Result:
        results = self._execute([Call(procedure=procedure, parameters=parameters.model_dump())])
        if len(results) != 1:
            raise RPCError(
                Error(
                    code=ErrorCode.ExecuteError,
                    message=f"Expected 1 result, got {len(results)}",
                )
            )

        if results[0].error and results[0].error.code != ErrorCode.ProcedureError:
            raise RPCError(results[0].error)

        return results[0]

    def execute_batch(self, procedures: List[Tuple[Procedure, Parameters]]) -> List[Result]:
        calls = [Call(procedure=procedure, parameters=parameters.model_dump()) for procedure, parameters in procedures]

        results = self._execute(calls)
        if len(results) != len(procedures):
            raise RPCError(
                Error(
                    code=ErrorCode.ExecuteError,
                    message=f"Expected {len(procedures)} results, got {len(results)}",
                )
            )

        return results
